// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
{**
  All types of ORB Exceptions.
}
unit exceptions;

interface

uses
  except_int,code_int,classes,types,sysutils,ORB_int;

type


  ORBException = class(TInterfacedObject,IORBException)
  protected
    procedure encode(const enc: IEncoder); virtual;
    procedure encode_any(const a: IAny); virtual;
    procedure print(s: TStream);
    function repoid(): PChar; virtual;
    function clone(): IORBException; virtual;
    procedure throw(); virtual;
    function ref_count(): integer;
  public
    class function decode(dec: IDecoder): IORBException;
    class function decode_any(a: IAny): IORBException;
  end;

  SystemException = class(ORBException,IORBException,ISystemException)
  protected
    {** current exception state }
    FStatus : TCompletionStatus;
    FMinor : _ulong;
    FSysExceptionType : SYSEXC_TYPES;
    function minor(): _ulong;
    function completed(): TCompletionStatus;
    procedure set_minor(val: _ulong);
    procedure set_completed(val: TCompletionStatus);
    procedure assign(const ex: ISystemException);
    function repoid(): PChar; override;
    function clone(): IORBException; override;
    function extype(): SYSEXC_TYPES;
  public
    constructor Create(ex: SystemException = nil);
    constructor Create_val(minor: _ulong; status: TCompletionStatus);
    constructor Create_type(extype: SYSEXC_TYPES; minor: _ulong = 0; status: TCompletionStatus = COMPLETED_YES);
    class function decode(dec: IDecoder): IORBException;
    class function decode_any(a: IAny): IORBException;
    class function create_sysex(repoid: PChar; minor: _ulong; stat: TCompletionStatus ): ISystemException;
    class function is_sysex(repoid: PChar): Boolean;
    class function EXCEPTION_DOWNCAST(const ex: IORBException): ISystemException;
  end;

  UserException = class(ORBException,IUserException)
  public
    class function decode(dec: IDecoder): IORBException;
    class function decode_any(a: IAny): IORBException;
  end;

  UnknownUserException = class(UserException,IUnknownUserException)
  private
    FDecoder : IDecoder;
    FStaticException : IStaticAny;
    FException: IAny;
    FEx_repoid : string;
  protected
    function exception(tc: ITypeCode): IAny; overload;
    function exception(sa: IStaticTypeInfo): IStaticAny; overload;
    function repoid(): PChar; override;
    function except_repoid: string;
  public
    constructor Create(dc: IDecoder);
    class function EXCEPTION_DOWNCAST(const ex: IORBException): IUnknownUserException;
  end;

  BadKind = class(UserException,IBadKind)
  protected
    function repoid(): PChar; override;
  public
    class function EXCEPTION_DOWNCAST(const ex: IORBException): IBadKind;
  end;

  Bounds = class(UserException,IBounds)
  protected
    function repoid(): PChar; override;
  public
    class function EXCEPTION_DOWNCAST(const ex: IORBException): IBounds;
  end;


implementation

uses
  static;

//************************************************************************
// TORBException
//************************************************************************
class function ORBException.decode(dec: IDecoder): IORBException;
var
  e: IORBException;
  pos: _ulong;
begin
  result := nil;
  pos := dec.buffer.RPosition;
  e := SystemException.decode(dec);
  if e = nil then
    begin
      dec.Buffer.RseekFromBegin(pos);
      e := UserException.decode(dec);
      if e = nil then
        begin
          dec.Buffer.RseekFromBegin(pos);
          exit;
        end;
    end;
  result := e;
end;

class function ORBException.decode_any(a: IAny): IORBException;
begin
  result := SystemException.decode_any(a);
  if result <> nil then exit;
  result := UserException.decode_any(a);
end;

function ORBException.ref_count(): integer;
begin
  result := FRefcount;
end;

function ORBException.clone(): IORBException;
begin
  assert(false);
  result := ORBException.Create();
end;

procedure ORBException.throw;
begin
  raise ORBException.Create;
end;

procedure ORBException.encode(const enc: IEncoder);
begin
  assert(false,'not implemented');
end;

procedure ORBException.encode_any(const a: IAny);
begin
  assert(false,'not implemented');
end;

procedure ORBException.print(s: TStream);
begin
  s.Write(repoid()^,length(repoid()+1));
end;

function ORBException.repoid: PChar;
begin
  result := 'IDL:omg.org/CORBA/Exception:1.0';
end;

//*********************************************************************
// TSystemException
//*********************************************************************
constructor SystemException.Create(ex: SystemException);
begin
  FSysExceptionType := st_SYSTEM;
  if ex = nil then
    begin
      FStatus := COMPLETED_NO;
      FMinor := 0;
    end
  else
    begin
      FStatus := ex.completed;
      FMinor := ex.minor;
    end;
end;

constructor SystemException.Create_val(minor: _ulong;
  status: TCompletionStatus);
begin
  FSysExceptionType := st_SYSTEM;
  FMinor := minor;
  FStatus := status;
end;

constructor SystemException.Create_type(extype: SYSEXC_TYPES; minor: _ulong; status: TCompletionStatus);
begin
  FSysExceptionType := extype;
  FMinor := minor;
  FStatus := status;
end;

procedure SystemException.assign(const ex: ISystemException);
begin
  FStatus := ex.completed;
  FMinor := ex.minor;
end;

function SystemException.completed(): TCompletionStatus;
begin
  result := FStatus;
end;

function SystemException.repoid(): PChar;
begin
  result := PChar('IDL:omg.org/CORBA/'+SYSEXC_NAMES[FSysExceptionType]+':1.0')
end;

function SystemException.clone(): IORBException;
begin
  result := SystemException.Create(self);
end;

class function SystemException.is_sysex(repoid: PChar): Boolean;
var
  i : SYSEXC_TYPES;
begin
  result := true;
  for i := st_SYSTEM to High(SYSEXC_TYPES) do
    if strcomp(repoid,PChar('IDL:omg.org/CORBA/'+SYSEXC_NAMES[i]+':1.0')) = 0 then
      exit;
  result := false;
end;

class function SystemException.EXCEPTION_DOWNCAST(const ex: IORBException): ISystemException;
begin
  result := nil;
  if is_sysex(ex.repoid) then
    result := ISystemException(ex)
end;

function SystemException.minor: _ulong;
begin
  result := FMinor;
end;

procedure SystemException.set_completed(val: TCompletionStatus);
begin
  FStatus := val;
end;

procedure SystemException.set_minor(val: _ulong);
begin
  FMinor := val;
end;

class function SystemException.decode(dec: IDecoder): IORBException;
var
  minor,val: _ulong;
  stat : TCompletionStatus;
  repoid: string;
begin
  result := nil;
  if not dec.get_string(repoid) or
     not dec.get_ulong(minor) or
     not dec.get_ulong(val) then exit;
  stat := TCompletionStatus(val);
  result := create_sysex(PChar(repoid),minor,stat);
end;

class function SystemException.decode_any(a: IAny): IORBException;
begin
  result := nil;
  assert(false,'not implemented');
end;

class function SystemException.create_sysex(repoid: PChar; minor: _ulong; stat: TCompletionStatus ): ISystemException;
var
  i : SYSEXC_TYPES;
begin
  result := nil;
  for i := st_SYSTEM to High(SYSEXC_TYPES) do
    if strcomp(repoid,PChar('IDL:omg.org/CORBA/'+SYSEXC_NAMES[i]+':1.0')) = 0 then
      begin
        result := SystemException.Create_type(i,minor,stat);
        break;
      end;
end;

function SystemException.extype(): SYSEXC_TYPES;
begin
  result := FSysExceptionType;
end;

//*********************************************************************
//  TUserException
//*********************************************************************
class function UserException.decode(dec: IDecoder): IORBException;
begin
  result := UnknownUserException.Create(dec);
end;

class function UserException.decode_any(a: IAny): IORBException;
begin
  assert(false,'not implemented');
end;

//*********************************************************************
//  TUnknownUserException
//*********************************************************************
constructor UnknownUserException.Create(dc: IDecoder);
begin
  FStaticException := nil;
  FException := nil;
  FDecoder := dc;
end;

function UnknownUserException.except_repoid: string;
var
  rpos : integer;
begin
  if FException <> nil then
    begin
      assert(false,'not implemented');
    end
  else
    begin
      rpos := FDecoder.buffer.RPosition;
      FDecoder.except_begin(FEx_repoid);
      FDecoder.buffer.RseekFromBegin(rpos);
    end;
  result := FEx_repoid;
end;

function UnknownUserException.repoid: PChar;
begin
  result := 'IDL:omg.org/CORBA/UnknownUserException:1.0';
end;

function UnknownUserException.exception(tc: ITypeCode): IAny;
begin
  assert(false,'not implemented');
end;

function UnknownUserException.exception(sa: IStaticTypeInfo): IStaticAny;
var
  rpos : integer;
  ret : Boolean;
begin
  if Assigned(FStaticException) then
    begin
      result := FStaticException;
      exit;
    end;
  FStaticException := TStaticAny.Create(sa);
  if assigned(FException) then
    begin

    end
  else
    begin
      rpos := FDecoder.buffer.RPosition;
      assert(assigned(FDecoder));
      ret := FStaticException.demarshal(FDecoder);
      assert(ret);
      FDecoder.buffer.RseekFromBegin(rpos);
    end;
  result := FStaticException;
end;

class function UnknownUserException.EXCEPTION_DOWNCAST(
  const ex: IORBException): IUnknownUserException;
begin
  result := nil;
  if not (ex = nil) and (strcomp(ex.repoid,'IDL:omg.org/CORBA/UnknownUserException:1.0')=0) then
    result := ex as IUnknownUserException;
end;

//*********************************************************************
//  BadKind
//*********************************************************************
class function BadKind.EXCEPTION_DOWNCAST(
  const ex: IORBException): IBadKind;
begin
  result := nil;
  if assigned(ex) and (ex.repoid = 'IDL:omg.org/CORBA/TypeCode/BadKind:1.0') then
    result := ex as IBadKind;
end;

function BadKind.repoid: PChar;
begin
  result := 'IDL:omg.org/CORBA/TypeCode/BadKind:1.0';
end;

//*********************************************************************
//  Bounds
//*********************************************************************
class function Bounds.EXCEPTION_DOWNCAST(const ex: IORBException): IBounds;
begin
  result := nil;
  if assigned(ex) and (ex.repoid = 'IDL:omg.org/CORBA/TypeCode/Bounds:1.0') then
    result := ex as IBounds;
end;

function Bounds.repoid: PChar;
begin
  result := 'IDL:omg.org/CORBA/TypeCode/Bounds:1.0'
end;

end.
